#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef MPI
#include "mpi.h"
#define MPI 1
#endif

#include "./MPI_Example_2.h"

int proc_MPI_msgs(MPI_Comm intracomm, MPI_Comm intercomm, int is_client)
{
    // basic storage for a few things to send around
    int dest_rank;
    char text_buf[6][256];
    
    // array of requests for MPI to use to employ nonblocking comms
    MPI_Request comms_reqs[6];
    MPI_Status comms_sts[6];

    MPI_Comm n_comm;
    
    // MPI error to return
    int err = MPI_SUCCESS;

    // are we in a multi-group scenario?
    if (intercomm == MPI_COMM_NULL) n_comm = &intracomm;
    else n_comm = &intercomm;
    //-------------------------------------------------------------------------------------------------------
    // No. We can safely assume intracomm is equivalent to MPI_COMM_WORLD.
    // Now do some simple rank-dependent functionality. This is rank 1.
    dest_rank = 0;
    // set up some nonblocking receives.
    if (err = MPI_Irecv(text_buf[2], 256, MPI_CHAR, dest_rank, 1, intracomm, &comms_reqs[2])) return err;
    if (err = MPI_Irecv(text_buf[3], 256, MPI_CHAR, dest_rank, 1, n_comm, &comms_reqs[3])) return err;
    //-----------------------------------------------------------------------------------------------------
    if (err = MPI_Waitall(2, &comms_reqs[2], &comms_sts[2])) return err;
    //-----------------------------------------------------------------------------------------------------
    // set up some nonblocking sends.
    if (err = MPI_Isend(text_buf[0], strlen(text_buf[1])+1, MPI_CHAR, dest_rank, 1, intracomm, &comms_reqs[0])) return err;
    if (err = MPI_Isend(text_buf[1], strlen(text_buf[2])+1, MPI_CHAR, dest_rank, 1, n_comm, &comms_reqs[1])) return err;    
    //-----------------------------------------------------------------------------------------------------
    if (err = MPI_Waitall(2, comms_reqs, comms_sts)) return err;
    //-----------------------------------------------------------------------------------------------------
    return err;
    //------------------------------------------------------------------------------------------------------
}

int main(int argc, char **argv)
{
    int err, client;
    MPI_Comm local_comm = MPI_COMM_NULL;
    MPI_Comm intergroup_comm = MPI_COMM_NULL;
    char port[MPI_MAX_PORT_NAME];

    MPI_Init(&argc, &argv);
    MPI_Comm_set_errhandler(MPI_COMM_WORLD, MPI_ERRORS_RETURN);
    if (NUM_GROUPS == 1) local_comm = MPI_COMM_WORLD;
    else
    {
       if (argc < 2)
       {
	  printf("Error: missing arg: client | server\n");
	  MPI_Finalize();
	  return 1;
       }
       if (!strcmp("client", argv[1]))
       {
	  if (err = MPI_Lookup_name(SRV_MPI_NAME, MPI_INFO_NULL, port))
	  {
	     printf("Error: no remote server found to connect to\n");
	     MPI_Finalize();
	     return err;
	  }
	  MPI_Comm_dup(MPI_COMM_WORLD, &local_comm);
	  MPI_Comm_set_errhandler(local_comm, MPI_ERRORS_RETURN);
	  if (err = MPI_Comm_connect(port, MPI_INFO_NULL, 0, local_comm, &intergroup_comm))
	  {
	     printf("Error: timed out trying to connect to remote server MPI_SRV_POETS at port %s", port);
	     MPI_Finalize();
	     return err;
	  }
	  client = 1;
       }
       else
       {
	  if (strcmp("server", argv[1]))
	  {
	     printf("Error: invalid command arg %s, must be one of {client, server}\n", argv[0]);
	     MPI_Finalize();
	     return 2;
	  }
	  MPI_Comm_dup(MPI_COMM_WORLD, &local_comm);
	  MPI_Comm_set_errhandler(local_comm, MPI_ERRORS_RETURN);
	  while (err = MPI_Lookup_name(SRV_MPI_NAME, MPI_INFO_NULL, port));
	  if (err = MPI_Comm_accept(port, MPI_INFO_NULL, 0, local_comm, &intergroup_comm)) MPI_Abort(local_comm, err);
	  client = 0;
       }
    }
    err = proc_MPI_msgs(local_comm, intergroup_comm, client);
    MPI_Finalize();
    return err;  	 
}
